/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2008,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _EXP2F4_H_
#define _EXP2F4_H_	1


#include <altivec.h>
#include <simdmath.h>

/*
 * FUNCTION
 *	vector float _exp2f4(vector float x)
 *
 * DESCRIPTION
 *	The _exp2f4 function computes 2 raised to the input vector x. 
 *      Computation is performed by observing the 2^(a+b) = 2^a * 2^b.
 *	We decompose x into a and b (above) by letting.
 *	a = ceil(x), b = x - a; 
 *
 *	2^a is easilty computed by placing a into the exponent
 *	or a floating point number whose mantissa is all zeros.
 *
 *	2^b is computed using the following polynomial approximation.
 *	(C. Hastings, Jr, 1955).
 *
 *                __7__
 *		  \
 *		   \ 
 *	2^(-x) =   /     Ci*x^i
 *                /____
 *                 i=1
 *
 *	for x in the range 0.0 to 1.0
 *
 *	C0 =  1.0
 *	C1 = -0.9999999995
 *	C2 =  0.4999999206
 *	C3 = -0.1666653019
 *	C4 =  0.0416573475
 *	C5 = -0.0083013598
 *	C6 =  0.0013298820
 *	C7 = -0.0001413161
 *
 */
static __inline vector float _exp2f4(vector float x)
{
  vector signed int ix;
  vector unsigned int overflow, underflow;
  vector float frac, frac2, frac4;
  vector float exp_int, exp_frac;
  vector float result;
  vector float hi, lo;
  vector float ceil;
  vector float vzero = ((vector float) { 0.0f, 0.0f, 0.0f, 0.0f });

  /* Break in the input x into two parts ceil(x), x - ceil(x).
   */
  ceil = vec_ceil(x);
  ix = vec_cts(ceil, 0);
  frac = vec_sub(ceil, x);
  frac = vec_madd(frac, ((vector float) { SM_LN2, SM_LN2, SM_LN2, SM_LN2 }), vzero);

  overflow = vec_sr((vector unsigned int)vec_cmpgt(ix, ((vector signed int) { 128, 128, 128, 128 })), ((vector unsigned int) { 1, 1, 1, 1 }));
  underflow = (vector unsigned int)vec_cmpgt(ix, ((vector signed int) { -128, -128, -128, -128 }));
  exp_int = (vector float)vec_and(vec_sl((vector unsigned int)vec_add(ix, ((vector signed int) { 127, 127, 127, 127 })), 
					 ((vector unsigned int) { 23, 23, 23, 23 })), underflow);

  /* Instruction counts can be reduced if the polynomial was
   * computed entirely from nested (dependent) fma's. However, 
   * to reduce the number of pipeline stalls, the polygon is evaluated 
   * in two halves (hi amd lo). 
   */
  frac2 = vec_madd(frac, frac, vzero);

  frac4 = vec_madd(frac2, frac2, vzero);
  hi = vec_madd(frac, ((vector float) { -0.0001413161f, -0.0001413161f, -0.0001413161f, -0.0001413161f }), ((vector float) { 0.0013298820f, 0.0013298820f, 0.0013298820f, 0.0013298820f }));
  hi = vec_madd(frac, hi, ((vector float) { -0.0083013598f, -0.0083013598f, -0.0083013598f, -0.0083013598f }));
  hi = vec_madd(frac, hi, ((vector float) { 0.0416573475f, 0.0416573475f, 0.0416573475f, 0.0416573475f }));
  lo = vec_madd(frac, ((vector float) { -0.1666653019f, -0.1666653019f, -0.1666653019f, -0.1666653019f }), ((vector float) { 0.4999999206f, 0.4999999206f, 0.4999999206f, 0.4999999206f }));
  lo = vec_madd(frac, lo, ((vector float) { -0.9999999995f, -0.9999999995f, -0.9999999995f, -0.9999999995f }));
  lo = vec_madd(frac, lo, ((vector float) { 1.0f, 1.0f, 1.0f, 1.0f }));

  exp_frac = vec_madd(frac4, hi, lo);

  ix = vec_add(ix, vec_sr((vector signed int)exp_frac, ((vector unsigned int) { 23, 23, 23, 23 })));
  result = vec_madd(exp_frac, exp_int, vzero);

  /* Handle overflow */
  result = vec_or(result, (vector float)(overflow));

  return (result);
}

#endif /* _EXP2F4_H_ */
