/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _ISEQUALF4_H_
#define _ISEQUALF4_H_	1

#include <altivec.h>

/*
 * FUNCTION
 *      vector unsigned int _isequalf4(vector float x, vector float y)
 *
 * DESCRIPTION
 *      The _isequalf4 function returns a vector in which each element 
 *      indicates if the corresponding element of x is equal to the 
 *      corresponding element of y.
 *     
 *      - The values +0 and -0 are considered equal.
 *      - If both inputs are infinity with the same sign, the comparison is 
 *        equal.
 *      - If either input is NaN, the comparison is false.
 *      - Denormals (subnormals) are compared correctly.
 *
 * RETURNS
 *       UINT_MAX (0xFFFFFFFF)  if the element of x is equal to the element of 
 *                              y.
 *       0        (0x00000000)  otherwise
 *
 */
static __inline vector unsigned int _isequalf4(vector float x, vector float y)
{
  vector unsigned int first_denorm = (vector unsigned int) { 0x7F800001,0x7F800001,0x7F800001,0x7F800001};

  vector unsigned int xabs         = (vector unsigned int) vec_abs(x);
  vector unsigned int yabs         = (vector unsigned int) vec_abs(y);

  vector unsigned int zero         = vec_xor(first_denorm,first_denorm);
  vector unsigned int equal        = (vector unsigned int)vec_cmpeq((vector unsigned int) x, (vector unsigned int) y);

  vector unsigned int x_not_a_nan  = (vector unsigned int)vec_cmplt(xabs,first_denorm); /* true if not a NaN */
  vector unsigned int y_not_a_nan  = (vector unsigned int)vec_cmplt(yabs,first_denorm); /* true if not a NaN */

  vector unsigned int zeros        = (vector unsigned int)vec_cmpeq(zero,vec_or(xabs,yabs)); /* check if x&y are both +/- zero */
   
  return (vector unsigned int)vec_or(zeros,	                                   /* Return ones if x&y are zero OR */
				     vec_and(equal,                                /* they're equal and */
					     vec_and(x_not_a_nan, y_not_a_nan)));  /* not NaNs */
}

#endif /* _ISEQUALF4_H_ */

