/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _TANHF4_H_
#define _TANHF4_H_	1

#include "expf4.h"
#include "divf4.h"

/*
 * Taylor coefficients for tanh
 */
#define TANH_TAY01   1.000000000000000000000000000000E0
#define TANH_TAY02  -3.333333333333333333333333333333E-1
#define TANH_TAY03   1.333333333333333333333333333333E-1
#define TANH_TAY04  -5.396825396825396825396825396825E-2
#define TANH_TAY05   2.186948853615520282186948853616E-2
#define TANH_TAY06  -8.863235529902196568863235529902E-3
#define TANH_TAY07   3.592128036572481016925461369906E-3
#define TANH_TAY08  -1.455834387051318268249485180702E-3
#define TANH_TAY09   5.900274409455859813780759937000E-4
#define TANH_TAY10  -2.391291142435524814857314588851E-4
#define TANH_TAY11   9.691537956929450325595875000389E-5
#define TANH_TAY12  -3.927832388331683405337080809312E-5
#define TANH_TAY13   1.591890506932896474074427981657E-5
#define TANH_TAY14  -6.451689215655430763190842315303E-6
#define TANH_TAY15   2.614771151290754554263594256410E-6
#define TANH_TAY16  -1.059726832010465435091355394125E-6
#define TANH_TAY17   4.294911078273805854820351280397E-7


/*
 * FUNCTION
 *  vector float _tanhf4(vector float x)
 *
 * DESCRIPTION
 *	The _tanhf4 function computes the hyperbolic tangent for each
 *	element of the input vector. 
 *
 *	We use the following to approximate tanh:
 *
 *	|x| <= .25:   Taylor Series
 *	|x| >  .25:   tanh(x) = (exp(2x) - 1)/(exp(2x) + 1)
 *
 *
 * SPECIAL CASES:
 *  - tanh(+/- 0) = +/-0
 *  - tanh(+/- infinity) = +/- 1
 *
 */

static __inline vector float _tanhf4(vector float x)
{
    vector float signbit = (vector float){-0.0f, -0.0f, -0.0f, -0.0f};
    vector float zerof   = (vector float){ 0.0f,  0.0f,  0.0f,  0.0f};
    vector float onef    = (vector float){ 1.0f,  1.0f,  1.0f,  1.0f};
    vector float twof    = (vector float){ 2.0f,  2.0f,  2.0f,  2.0f};
    vector float xabs;
    vector float x2;
    vector bool  int gttaylor;
    vector float e;
    vector float tresult;
    vector float eresult;
    vector float result;

    xabs = vec_andc(x, signbit);

    /*
     * This is where we switch from Taylor Series
     * to exponential formula.
     */
    gttaylor = vec_cmpgt(xabs, ((vector float){0.25f, 0.25f, 0.25f, 0.25f}));


    /*
     * Taylor Series Approximation
     */
    x2 = vec_madd(x,x, zerof);
    tresult = vec_madd(x2, ((vector float){TANH_TAY06, TANH_TAY06, TANH_TAY06, TANH_TAY06}), 
                           ((vector float){TANH_TAY05, TANH_TAY05, TANH_TAY05, TANH_TAY05}));
    tresult = vec_madd(x2, tresult, ((vector float){TANH_TAY04, TANH_TAY04, TANH_TAY04, TANH_TAY04}));
    tresult = vec_madd(x2, tresult, ((vector float){TANH_TAY03, TANH_TAY03, TANH_TAY03, TANH_TAY03}));
    tresult = vec_madd(x2, tresult, ((vector float){TANH_TAY02, TANH_TAY02, TANH_TAY02, TANH_TAY02}));
    tresult = vec_madd(x2, tresult, ((vector float){TANH_TAY01, TANH_TAY01, TANH_TAY01, TANH_TAY01}));
    tresult = vec_madd(xabs, tresult, zerof);


    /*
     * Exponential Formula
     * Our expf4 function gives a more accurate result in general 
     * with xabs instead of x for x<0. We correct for sign later.
     */
    e = _expf4(vec_madd(xabs, twof, zerof));
    eresult = _divf4(vec_sub(e, onef), vec_add(e, onef));


    /*
     * Select Taylor or exp result.
     */
    result = vec_sel(tresult, eresult, gttaylor);

    /*
     * Correct for accumulated truncation error when 
     * tanh(x) should return 1.
     * Note that this also handles the special case of 
     * x = +/- infinity.
     */
    result = vec_sel(result, onef, vec_cmpgt(xabs, ((vector float){9.125f, 9.125f, 9.125f, 9.125f})));

    /*
     * Antisymmetric function - preserve sign bit of x
     * in the result.
     */
    result = vec_sel(result, x, (vec_uint4)signbit);

    return result;
}

#endif /* _TANHF4_H_ */
